@@ -17,7 +17,7 @@ def array2compressed(array, quality=90):
     """Convert the given image (a numpy array) as a compressed array.
 
     If the quality is 100, a PNG is returned. Otherwise, JPEG is
-    preferred and PNG is used as a fallback. Returns (preamble, bytes).
+    preferred and PNG is used as a fallback. Returns (mimetype, bytes).
     """
 
     # Drop alpha channel if there is one
@@ -25,16 +25,16 @@ def array2compressed(array, quality=90):
         array = array[:, :, :3]
 
     if quality >= 100:
-        preamble = "data:image/png;base64,"
+        mimetype = "image/png"
         result = array2png(array)
     else:
-        preamble = "data:image/jpeg;base64,"
+        mimetype = "image/jpeg"
         result = array2jpg(array, quality)
         if result is None:
-            preamble = "data:image/png;base64,"
+            mimetype = "image/png"
             result = array2png(array)
 
-    return preamble, result
+    return mimetype, result
 
 
 class RFBOutputContext(ipywidgets.Output):
@@ -117,3 +117,32 @@ class Snapshot(DisplayObject):
             </div>
             """
         return html.replace("\n", "").replace("    ", "").strip()
+
+
+def remove_rfb_models_from_nb(d):
+    """Remove the widget model output from a notebook dict.
+
+    Given a notebook as a dict (loaded using json), remove the widget
+    model output if there is also a text/html snapshot output.
+
+    This is to work around the fact that nbsphinx favors the model over
+    the text/html output. Which is sad, because that's where we put the
+    initial screenshot for offline viewing.
+    """
+
+    to_remove = set()
+    for key, val in d.items():
+        if key == "cells" and isinstance(val, list):
+            for v in val:
+                remove_rfb_models_from_nb(v)
+        elif key == "outputs" and isinstance(val, list):
+            for v in val:
+                data = v.get("data", None)
+                if data:
+                    remove_rfb_models_from_nb(data)
+        elif key == "application/vnd.jupyter.widget-view+json":
+            html_sibling = d.get("text/html", [])
+            if html_sibling and "<div class='snapshot-" in html_sibling[0]:
+                to_remove.add(key)
+    for key in to_remove:
+        d.pop(key)
